"""Asynchronous Python client for Russound RIO."""


class RussoundError(Exception):
    """A generic error."""


class CommandError(Exception):
    """A command sent to the controller caused an error."""


class UncachedVariableError(Exception):
    """A variable was not found in the cache."""


class UnsupportedFeatureError(Exception):
    """A requested command is not supported on this controller."""


class UnsupportedRussoundVersionError(Exception):
    """The client implements an unsupported version of the Russound RIO API."""



"""Asynchronous Python client for Russound RIO."""

from __future__ import annotations

from collections import defaultdict
from enum import Enum
import re

MINIMUM_API_SUPPORT = "1.05.00"

DEFAULT_PORT = 9621

RECONNECT_DELAY = 5.0
TIMEOUT = 5.0

KEEP_ALIVE_INTERVAL = 60

MAX_SOURCE = 17

MAX_RNET_CONTROLLERS = 6

RESPONSE_REGEX = re.compile(
    r'(?:(\w+(?:\[\d+])?(?:\.\w+(?:\[\d+])?)*)\.)?(\w+)="([^"]*)"'
)


class FeatureFlag(Enum):
    """A list of Russound RIO API features."""

    SUPPORT_POWER_MGMT = 1
    SUPPORT_ZONE_PARAMETERS = 2
    SUPPORT_SOURCE_PARAMETERS = 3
    SUPPORT_CONTROLLER_PARAMETERS = 4
    SUPPORT_SYSTEM_PARAMETERS = 5
    SUPPORT_DMS_3_1_MM = 6
    EVENT_KEY_CODE = 7
    PROPERTY_IP_ADDRESS = 8
    SUPPORT_SHUFFLE = 9
    COMMAND_MM_CLOSE = 10
    SUPPORT_PAGE_ZONE = 11
    SUPPORT_REPEAT = 12
    PROPERTY_CTRL_TYPE = 13
    PROPERTY_SYS_LANG = 14
    NOTIFICATION_SYS_LANG = 15
    SUPPORT_MM_LONG_LIST = 16
    TEMPLATES_MM_SCREEN = 17
    PROPERTY_SLEEP_TIME_REMAINING = 18
    SUPPORT_PRESETS_BANKS = 19
    SUPPORT_FAVORITES = 20
    SUPPORT_ZONE_SOURCE_EXCLUSION = 21
    SUPPORT_FORMS = 22
    SUPPORT_MEDIA_RATING = 23
    SUPPORT_HIDDEN_ATTRIBUTE = 24
    SUPPORT_SYSTEM_FAVORITE_RENAME = 25
    COMMANDS_ZONE_MUTE_OFF_ON = 26
    SUPPORT_WATCH_FAVORITES = 27
    SUPPORT_SYSTEM = 28
    SUPPORT_DEVICE_GROUPING = 29
    SUPPORT_ALARM = 30
    NOTIFICATION_ALARM_ZONE_WATCH = 31
    PROPERTY_FIRMWARE_VERSION = 32
    SUPPORT_MBX_DISPLAY_ITEMS = 33
    PROPERTY_PLAY_STATUS = 34
    PROPERTY_AVAILABLE_CONTROLS = 35
    PROPERTY_SAMPLE_RATE = 36
    PROPERTY_BIT_RATE = 37
    PROPERTY_BIT_DEPTH = 38
    PROPERTY_PLAY_TIME = 39
    PROPERTY_TRACK_TIME = 40
    PROPERTY_SET_SEEK_TIME = 41
    SUPPORT_MM_CONTEXT_MENU = 42
    PROPERTY_SLEEP_TIME_DEFAULT = 43
    PROPERTY_SUPPORT_SLEEP_TIME = 44
    EVENT_REBOOT = 45
    SUPPORT_SYSTEM_FAVORITE_SOURCE = 46
    ATTRIBUTE_USER_LOGIN = 47


FLAGS_BY_VERSION = {
    "1.01.00": [
        FeatureFlag.SUPPORT_POWER_MGMT,
    ],
    "1.02.00": [
        FeatureFlag.SUPPORT_ZONE_PARAMETERS,
        FeatureFlag.SUPPORT_SOURCE_PARAMETERS,
        FeatureFlag.SUPPORT_CONTROLLER_PARAMETERS,
        FeatureFlag.SUPPORT_SYSTEM_PARAMETERS,
    ],
    "1.03.00": [
        FeatureFlag.SUPPORT_DMS_3_1_MM,
        FeatureFlag.EVENT_KEY_CODE,
        FeatureFlag.PROPERTY_IP_ADDRESS,
    ],
    "1.04.00": [
        FeatureFlag.SUPPORT_SHUFFLE,
        FeatureFlag.COMMAND_MM_CLOSE,
        FeatureFlag.SUPPORT_PAGE_ZONE,
    ],
    "1.05.00": [
        FeatureFlag.SUPPORT_REPEAT,
        FeatureFlag.PROPERTY_CTRL_TYPE,
        FeatureFlag.PROPERTY_SYS_LANG,
    ],
    "1.06.00": [
        FeatureFlag.NOTIFICATION_SYS_LANG,
        FeatureFlag.SUPPORT_MM_LONG_LIST,
        FeatureFlag.TEMPLATES_MM_SCREEN,
    ],
    "1.07.00": [
        FeatureFlag.PROPERTY_SLEEP_TIME_REMAINING,
        FeatureFlag.SUPPORT_PRESETS_BANKS,
        FeatureFlag.SUPPORT_FAVORITES,
        FeatureFlag.SUPPORT_ZONE_SOURCE_EXCLUSION,
        FeatureFlag.SUPPORT_FORMS,
        FeatureFlag.SUPPORT_MEDIA_RATING,
        FeatureFlag.SUPPORT_HIDDEN_ATTRIBUTE,
    ],
    "1.08.00": [FeatureFlag.SUPPORT_SYSTEM_FAVORITE_RENAME],
    "1.09.00": [
        FeatureFlag.COMMANDS_ZONE_MUTE_OFF_ON,
    ],
    "1.11.00": [
        FeatureFlag.SUPPORT_WATCH_FAVORITES,
    ],
    "1.12.00": [
        FeatureFlag.SUPPORT_SYSTEM,
        FeatureFlag.SUPPORT_DEVICE_GROUPING,
        FeatureFlag.SUPPORT_ALARM,
    ],
    "1.12.01": [
        FeatureFlag.NOTIFICATION_ALARM_ZONE_WATCH,
    ],
    "1.12.02": [
        FeatureFlag.PROPERTY_FIRMWARE_VERSION,
    ],
    "1.14.00": [
        FeatureFlag.SUPPORT_MBX_DISPLAY_ITEMS,
        FeatureFlag.PROPERTY_PLAY_STATUS,
        FeatureFlag.PROPERTY_AVAILABLE_CONTROLS,
        FeatureFlag.PROPERTY_SAMPLE_RATE,
        FeatureFlag.PROPERTY_BIT_RATE,
        FeatureFlag.PROPERTY_BIT_DEPTH,
        FeatureFlag.PROPERTY_PLAY_TIME,
        FeatureFlag.PROPERTY_TRACK_TIME,
        FeatureFlag.PROPERTY_SET_SEEK_TIME,
        FeatureFlag.SUPPORT_MM_CONTEXT_MENU,
    ],
    "1.14.01": [
        FeatureFlag.PROPERTY_SLEEP_TIME_DEFAULT,
        FeatureFlag.PROPERTY_SUPPORT_SLEEP_TIME,
    ],
    "1.15.00": [FeatureFlag.EVENT_REBOOT, FeatureFlag.SUPPORT_SYSTEM_FAVORITE_SOURCE],
    "1.15.02": [
        FeatureFlag.ATTRIBUTE_USER_LOGIN,
    ],
}

VERSIONS_BY_FLAGS = defaultdict(list)

for version, flags in FLAGS_BY_VERSION.items():
    for flag in flags:
        VERSIONS_BY_FLAGS[flag] = version

CONTROLLER_TYPE_FIX_MAP = {
    "MCA-C6": "MCA-C5",
    "XStream-X5": "XSource",
}


"""Asynchronous Python client for Russound RIO."""

from __future__ import annotations

import asyncio
import logging
from asyncio import Future, Task, AbstractEventLoop, Queue
from dataclasses import field, dataclass
from typing import Any, Coroutine, Optional

from aiorussound.connection import RussoundConnectionHandler
from aiorussound.const import (
    FLAGS_BY_VERSION,
    MAX_SOURCE,
    MINIMUM_API_SUPPORT,
    FeatureFlag,
    MAX_RNET_CONTROLLERS,
    RESPONSE_REGEX,
    KEEP_ALIVE_INTERVAL,
    TIMEOUT,
    CONTROLLER_TYPE_FIX_MAP,
)
from aiorussound.exceptions import (
    CommandError,
    UnsupportedFeatureError,
    RussoundError,
)
from aiorussound.models import (
    RussoundMessage,
    CallbackType,
    Source,
    Zone,
    MessageType,
)
from aiorussound.util import (
    controller_device_str,
    is_feature_supported,
    is_fw_version_higher,
    source_device_str,
    zone_device_str,
    is_rnet_capable,
    get_max_zones,
    map_rio_to_dict,
)

_LOGGER = logging.getLogger(__package__)


class RussoundClient:
    """Manages the RIO connection to a Russound device."""

    def __init__(self, connection_handler: RussoundConnectionHandler) -> None:
        """Initialize the Russound object using the event loop, host and port
        provided.
        """
        self.connection_handler = connection_handler
        self._loop: AbstractEventLoop = asyncio.get_running_loop()
        self._subscriptions: dict[str, Any] = {}
        self.connect_result: Future | None = None
        self.connect_task: Task | None = None
        self._reconnect_task: Optional[Task] = None
        self._state_update_callbacks: list[Any] = []
        self.controllers: dict[int, Controller] = {}
        self.sources: dict[int, Source] = {}
        self.rio_version: str | None = None
        self.state = {}
        self._futures: Queue = Queue()
        self._attempt_reconnection = False
        self._do_state_update = False

    async def register_state_update_callbacks(self, callback: Any):
        """Register state update callback."""
        self._state_update_callbacks.append(callback)
        if self._do_state_update:
            await callback(self, CallbackType.STATE)

    def unregister_state_update_callbacks(self, callback: Any):
        """Unregister state update callback."""
        if callback in self._state_update_callbacks:
            self._state_update_callbacks.remove(callback)

    def clear_state_update_callbacks(self):
        """Clear state update callbacks."""
        self._state_update_callbacks.clear()

    async def do_state_update_callbacks(
        self, callback_type: CallbackType = CallbackType.STATE
    ):
        """Call state update callbacks."""
        if not self._state_update_callbacks:
            return
        callbacks = set()
        for callback in self._state_update_callbacks:
            callbacks.add(callback(self, callback_type))

        if callbacks:
            await asyncio.gather(*callbacks)

    async def request(self, cmd: str):
        _LOGGER.debug("Sending command '%s' to Russound client", cmd)
        future: Future = Future()
        await self._futures.put(future)
        try:
            await self.connection_handler.send(cmd)
        except Exception as ex:
            _ = await self._futures.get()
            future.set_exception(ex)
        return await future

    async def connect(self) -> None:
        """Connect to the controller and start processing responses."""
        if not self.is_connected():
            self.connect_result = self._loop.create_future()
            self._reconnect_task = asyncio.create_task(
                self._reconnect_handler(self.connect_result)
            )
        return await self.connect_result

    async def disconnect(self) -> None:
        """Disconnect from the Russound controller."""
        if self.is_connected():
            self._attempt_reconnection = False
            self.connect_task.cancel()
            try:
                await self.connect_task
            except asyncio.CancelledError:
                pass

    def is_connected(self) -> bool:
        """Return True if device is connected."""
        return self.connect_task is not None and not self.connect_task.done()

    async def _reconnect_handler(self, res):
        reconnect_delay = 0.5
        while True:
            try:
                self.connect_task = asyncio.create_task(self._connect_handler(res))
                await self.connect_task
            except Exception as ex:
                _LOGGER.error(ex)
                pass
            await self.do_state_update_callbacks(CallbackType.CONNECTION)
            if not self._attempt_reconnection:
                _LOGGER.debug(
                    "Failed to connect to device on initial pass, skipping reconnect."
                )
                break
            reconnect_delay = min(reconnect_delay * 2, 30)
            _LOGGER.debug(
                f"Attempting reconnection to Russound device in {reconnect_delay} seconds..."
            )
            await asyncio.sleep(reconnect_delay)

    async def _connect_handler(self, res):
        handler_tasks = set()
        try:
            self._do_state_update = False
            async with asyncio.timeout(TIMEOUT):
                await self.connection_handler.connect()
            handler_tasks.add(
                asyncio.create_task(self.consumer_handler(self.connection_handler))
            )
            self.rio_version = await self.request("VERSION")
            if not is_fw_version_higher(self.rio_version, MINIMUM_API_SUPPORT):
                raise UnsupportedFeatureError(
                    f"Russound RIO API v{self.rio_version} is not supported. The minimum "
                    f"supported version is v{MINIMUM_API_SUPPORT}"
                )
            _LOGGER.info("Connected (Russound RIO v%s})", self.rio_version)
            # Fetch parent controller
            parent_controller = await self._load_controller(1)
            if not parent_controller:
                raise RussoundError("No primary controller found.")

            self.controllers[1] = parent_controller

            # Only search for daisy-chained controllers if the parent supports RNET
            if is_rnet_capable(parent_controller.controller_type):
                for controller_id in range(2, MAX_RNET_CONTROLLERS + 1):
                    controller = await self._load_controller(controller_id)
                    if controller:
                        self.controllers[controller_id] = controller

            self._do_state_update = True
            self._attempt_reconnection = True
            if not res.done():
                res.set_result(True)
            handler_tasks.add(asyncio.create_task(self._keep_alive()))
            await asyncio.wait(handler_tasks, return_when=asyncio.FIRST_COMPLETED)
        except Exception as ex:
            if not res.done():
                res.set_exception(ex)
            _LOGGER.error(ex, exc_info=True)
        finally:
            for task in handler_tasks:
                if not task.done():
                    task.cancel()

            while not self._futures.empty():
                future = await self._futures.get()
                future.cancel()

            self._do_state_update = False

            closeout = set()
            closeout.update(handler_tasks)

            if closeout:
                closeout_task = asyncio.create_task(asyncio.wait(closeout))
                while not closeout_task.done():
                    try:
                        await asyncio.shield(closeout_task)
                    except asyncio.CancelledError:
                        pass

    async def load_zone_source_metadata(self) -> None:
        """Fetches and subscribes to all the zone and source metadata"""

        subscribe_state_updates = {self.subscribe(self._async_handle_system, "System")}

        # Load source structure
        for source_id in range(1, MAX_SOURCE):
            try:
                device_str = source_device_str(source_id)
                name = await self.get_variable(device_str, "name")
                if name:
                    subscribe_state_updates.add(
                        self.subscribe(self._async_handle_source, device_str)
                    )
            except CommandError:
                break

        for controller_id, controller in self.controllers.items():
            for zone_id in range(1, get_max_zones(controller.controller_type) + 1):
                try:
                    device_str = zone_device_str(controller_id, zone_id)
                    name = await self.get_variable(device_str, "name")
                    if name:
                        subscribe_state_updates.add(
                            self.subscribe(self._async_handle_zone, device_str)
                        )
                except CommandError:
                    break

        subscribe_tasks = set()
        for state_update in subscribe_state_updates:
            subscribe_tasks.add(asyncio.create_task(state_update))
        await asyncio.wait(subscribe_tasks)

        if is_feature_supported(
            self.rio_version, FeatureFlag.SUPPORT_ZONE_SOURCE_EXCLUSION
        ):
            _LOGGER.debug(
                "Zone source exclusion is supported. Fetching excluded sources."
            )
            await self._load_zone_source_exclusion()
            # Reload zones from state
            await self._async_handle_zone()

        await self.do_state_update_callbacks(CallbackType.STATE)

        # Delay to ensure async TTL
        await asyncio.sleep(0.5)

    @staticmethod
    def process_response(res: bytes) -> Optional[RussoundMessage]:
        """Process an incoming string of bytes into a RussoundMessage"""
        try:
            # Attempt to decode in Latin and re-encode in UTF-8 to support international characters
            str_res = (
                res.decode(encoding="iso-8859-1")
                .encode(encoding="utf-8")
                .decode(encoding="utf-8")
                .strip()
            )
        except UnicodeDecodeError as e:
            _LOGGER.warning("Failed to decode Russound response %s", res, e)
            return None
        if not str_res:
            return None
        if len(str_res) == 1 and str_res[0] == "S":
            return RussoundMessage(MessageType.STATE, None, None, None)
        tag, payload = str_res[0], str_res[2:]
        if tag == "E":
            _LOGGER.debug("Device responded with error: %s", payload)
            return RussoundMessage(tag, None, None, payload)
        m = RESPONSE_REGEX.match(payload.strip())
        if not m:
            return RussoundMessage(tag, None, None, None)
        value = m.group(3)
        value = None if not value or value == "------" else value
        return RussoundMessage(tag, m.group(1) or None, m.group(2), value)

    async def consumer_handler(self, handler: RussoundConnectionHandler):
        """Callback consumer handler."""
        try:
            async for raw_msg in handler.reader:
                msg = self.process_response(raw_msg)
                if msg:
                    _LOGGER.debug(f"recv ({msg})")
                    if msg.type == "S" and not self._futures.empty():
                        future: Future = await self._futures.get()
                        if not future.done():
                            future.set_result(msg.value)
                    elif msg.type == "E" and not self._futures.empty():
                        future: Future = await self._futures.get()
                        if not future.done():
                            future.set_exception(CommandError)
                    if msg.branch and msg.leaf and msg.type == "N":
                        map_rio_to_dict(self.state, msg.branch, msg.leaf, msg.value)
                        subscription = self._subscriptions.get(msg.branch)
                        if subscription:
                            await subscription()
        except (asyncio.CancelledError, OSError):
            pass

    async def _keep_alive(self) -> None:
        while True:
            await asyncio.sleep(KEEP_ALIVE_INTERVAL)
            _LOGGER.debug("Sending keep alive to device")
            try:
                async with asyncio.timeout(TIMEOUT):
                    await self.request("VERSION")
            except asyncio.TimeoutError:
                _LOGGER.warning("Keep alive request to the Russound device timed out")
                break
        _LOGGER.debug("Ending keep alive task to attempt reconnection")

    async def subscribe(self, callback: Any, branch: str) -> None:
        self._subscriptions[branch] = callback
        try:
            await self.request(f"WATCH {branch} ON")
        except (asyncio.CancelledError, asyncio.TimeoutError, CommandError):
            del self._subscriptions[branch]
            raise

    async def _async_handle_system(self) -> None:
        """Handle async info update."""
        if self._do_state_update:
            await self.do_state_update_callbacks()

    async def _async_handle_source(self) -> None:
        """Handle async info update."""
        for source_id, source_data in self.state["S"].items():
            source = Source.from_dict(source_data)
            source.client = self
            self.sources[source_id] = source
        if self._do_state_update:
            await self.do_state_update_callbacks()

    async def _async_handle_zone(self) -> None:
        """Handle async info update."""
        for controller_id, controller_data in self.state["C"].items():
            for zone_id, zone_data in controller_data["Z"].items():
                zone = ZoneControlSurface.from_dict(zone_data)
                zone.client = self
                zone.device_str = zone_device_str(controller_id, zone_id)
                self.controllers[controller_id].zones[zone_id] = zone
        if self._do_state_update:
            await self.do_state_update_callbacks()

    async def set_variable(
        self, device_str: str, key: str, value: str
    ) -> Coroutine[Any, Any, str]:
        """Set a zone variable to a new value."""
        return await self.request(f'SET {device_str}.{key}="{value}"')

    async def get_variable(self, device_str: str, key: str) -> str:
        """Retrieve the current value of a zone variable.  If the variable is
        not found in the local cache then the value is requested from the
        controller.
        """
        return await self.request(f"GET {device_str}.{key}")

    async def _load_controller(self, controller_id: int) -> Optional[Controller]:
        device_str = controller_device_str(controller_id)
        try:
            controller_type = await self.get_variable(device_str, "type")
            if not controller_type:
                return None
            if controller_type in CONTROLLER_TYPE_FIX_MAP:
                controller_type = CONTROLLER_TYPE_FIX_MAP[controller_type]
            mac_address = None
            try:
                mac_address = await self.get_variable(device_str, "macAddress")
            except CommandError:
                pass
            firmware_version = None
            if is_feature_supported(
                self.rio_version, FeatureFlag.PROPERTY_FIRMWARE_VERSION
            ):
                firmware_version = await self.get_variable(
                    device_str, "firmwareVersion"
                )
            controller = Controller(
                controller_id,
                controller_type,
                self,
                controller_device_str(controller_id),
                mac_address,
                firmware_version,
                {},
            )
            return controller
        except CommandError:
            return None

    # ----------------------
    # Manual state fixes
    # ----------------------

    async def _load_zone_source_exclusion(self) -> None:
        """Loads whether a source is available to a specific zone."""
        for controller_id, controller in self.controllers.items():
            for zone_id in controller.zones.keys():
                for source_id in self.sources.keys():
                    try:
                        enabled = await self.get_variable(
                            f"C[{controller_id}].Z[{zone_id}].S[{source_id}]", "enabled"
                        )
                    except CommandError:
                        continue
                    if enabled == "TRUE":
                        if (
                            "enabled_sources"
                            not in self.state["C"][controller_id]["Z"][zone_id]
                        ):
                            self.state["C"][controller_id]["Z"][zone_id][
                                "enabled_sources"
                            ] = []
                        self.state["C"][controller_id]["Z"][zone_id][
                            "enabled_sources"
                        ].append(source_id)

    @property
    def supported_features(self) -> list[FeatureFlag]:
        """Gets a list of features supported by the controller."""
        flags: list[FeatureFlag] = []
        for key, value in FLAGS_BY_VERSION.items():
            if is_fw_version_higher(self.rio_version, key):
                for flag in value:
                    flags.append(flag)
        return flags


class AbstractControlSurface:
    def __init__(self):
        self.client: Optional[RussoundClient] = None
        self.device_str: Optional[str] = None


class ZoneControlSurface(Zone, AbstractControlSurface):
    async def send_event(self, event_name, *args) -> str:
        """Send an event to a zone."""
        args = " ".join(str(x) for x in args)
        cmd = f"EVENT {self.device_str}!{event_name} {args}"
        return await self.client.request(cmd)

    def fetch_current_source(self) -> Source:
        """Return the current source as a source object."""
        return self.client.sources[self.current_source]

    async def mute(self) -> str:
        """Mute the zone."""
        return await self.send_event("ZoneMuteOn")

    async def unmute(self) -> str:
        """Unmute the zone."""
        return await self.send_event("ZoneMuteOff")

    async def toggle_mute(self) -> str:
        """Toggle the mute state of the zone."""
        return await self.send_event("KeyRelease", "Mute")

    async def set_volume(self, volume: str) -> str:
        """Set the volume."""
        return await self.send_event("KeyPress", "Volume", volume)

    async def volume_up(self) -> str:
        """Volume up the zone."""
        return await self.send_event("KeyPress", "VolumeUp")

    async def volume_down(self) -> str:
        """Volume down the zone."""
        return await self.send_event("KeyPress", "VolumeDown")

    async def previous(self) -> str:
        """Go to the previous song."""
        return await self.send_event("KeyPress", "Previous")

    async def next(self) -> str:
        """Go to the next song."""
        return await self.send_event("KeyPress", "Next")

    async def stop(self) -> str:
        """Stop the current song."""
        return await self.send_event("KeyPress", "Stop")

    async def pause(self) -> str:
        """Pause the current song."""
        return await self.send_event("KeyPress", "Pause")

    async def play(self) -> str:
        """Play the queued song."""
        return await self.send_event("KeyPress", "Play")

    async def zone_on(self) -> str:
        """Turn on the zone."""
        return await self.send_event("ZoneOn")

    async def zone_off(self) -> str:
        """Turn off the zone."""
        return await self.send_event("ZoneOff")

    async def select_source(self, source: int) -> str:
        """Select a source."""
        return await self.send_event("SelectSource", source)

    async def set_seek_time(self, time: int) -> None:
        """Seek to the specified time."""
        if time < 0:
            raise RussoundError("Seek time cannot be negative")
        elif time > self.fetch_current_source().track_time:
            raise RussoundError("Seek time cannot be greater than current track time")
        await self.send_event("SetSeekTime", time)

    async def set_loudness(self, loudness: bool) -> None:
        """Set the loudness of the zone."""
        await self.client.set_variable(
            self.device_str, "loudness", "ON" if loudness else "OFF"
        )

    async def set_bass(self, bass: int) -> None:
        """Set the bass of the zone."""
        if bass < -10 or bass > 10:
            raise RussoundError("Bass must be between -10 and 10")
        await self.client.set_variable(self.device_str, "bass", str(bass))

    async def set_treble(self, treble: int) -> None:
        """Set the treble of the zone."""
        if treble < -10 or treble > 10:
            raise RussoundError("Treble must be between -10 and 10")
        await self.client.set_variable(self.device_str, "treble", str(treble))

    async def set_balance(self, balance: int) -> None:
        """Set the treble of the zone."""
        if balance < -10 or balance > 10:
            raise RussoundError("Balance must be between -10 and 10")
        await self.client.set_variable(self.device_str, "balance", str(balance))

    async def set_turn_on_volume(self, turn_on_volume: int) -> None:
        """Set the turn on volume of the zone."""
        if turn_on_volume < 0 or turn_on_volume > 50:
            raise RussoundError("Turn on volume must be between 0 and 50")
        await self.client.set_variable(
            self.device_str, "turn_on_volume", str(turn_on_volume)
        )


@dataclass
class Controller:
    """Data class representing a Russound controller."""

    controller_id: int
    controller_type: str
    client: RussoundClient
    device_str: str
    mac_address: Optional[str]
    firmware_version: Optional[str]
    zones: dict[int, ZoneControlSurface] = field(default_factory=dict)



"""Models for aiorussound."""

from dataclasses import dataclass, field
from enum import StrEnum
from typing import Optional

from mashumaro import field_options
from mashumaro.mixins.orjson import DataClassORJSONMixin
from mashumaro.types import SerializationStrategy
from datetime import datetime, UTC


class RussoundBool(SerializationStrategy):
    def deserialize(self, value: str) -> bool:
        if value and (value == "ON" or value == "TRUE"):
            return True
        return False


class RussoundInt(SerializationStrategy):
    def deserialize(self, value: str) -> int:
        return int(value)


@dataclass
class Zone(DataClassORJSONMixin):
    """Data class representing Russound state."""

    name: str = field(default=None)
    volume: int = field(
        metadata=field_options(serialization_strategy=RussoundInt()), default=0
    )
    bass: int = field(
        metadata=field_options(serialization_strategy=RussoundInt()), default=0
    )
    treble: int = field(
        metadata=field_options(serialization_strategy=RussoundInt()), default=0
    )
    balance: int = field(
        metadata=field_options(serialization_strategy=RussoundInt()), default=0
    )
    loudness: bool = field(
        metadata=field_options(serialization_strategy=RussoundBool()), default=False
    )
    turn_on_volume: int = field(
        metadata=field_options(
            alias="turnOnVolume", serialization_strategy=RussoundInt()
        ),
        default=20,
    )
    do_not_disturb: bool = field(
        metadata=field_options(
            alias="doNotDisturb", serialization_strategy=RussoundBool()
        ),
        default=False,
    )
    party_mode: bool = field(
        metadata=field_options(
            alias="partyMode", serialization_strategy=RussoundBool()
        ),
        default=False,
    )
    status: bool = field(
        metadata=field_options(serialization_strategy=RussoundBool()), default=False
    )
    is_mute: bool = field(
        metadata=field_options(alias="mute", serialization_strategy=RussoundBool()),
        default=False,
    )
    shared_source: bool = field(
        metadata=field_options(
            alias="sharedSource", serialization_strategy=RussoundBool()
        ),
        default=False,
    )
    last_error: Optional[str] = field(
        metadata=field_options(alias="lastError"), default=None
    )
    page: Optional[str] = field(metadata=field_options(alias="page"), default=None)
    sleep_time_default: Optional[int] = field(
        metadata=field_options(
            alias="sleepTimeDefault", serialization_strategy=RussoundInt()
        ),
        default=None,
    )
    sleep_time_remaining: Optional[int] = field(
        metadata=field_options(
            alias="sleepTimeRemaining", serialization_strategy=RussoundInt()
        ),
        default=None,
    )
    enabled: bool = field(
        metadata=field_options(serialization_strategy=RussoundBool()), default=False
    )
    current_source: int = field(
        metadata=field_options(
            alias="currentSource", serialization_strategy=RussoundInt()
        ),
        default=1,
    )
    enabled_sources: list[int] = field(
        metadata=field_options(alias="enabled_sources"), default_factory=list
    )


class SourceType(StrEnum):
    """Russound source types."""

    AMPLIFIER = "Amplifier"
    TELEVISION = "Television"
    CABLE = "Cable"
    VIDEO_ACCESSORY = "Video Accessory"
    SATELLITE = "Satellite"
    VCR = "VCR"
    BLURAY_DVD = "Blu-ray / DVD"
    RECEIVER = "Receiver"
    MISC_AUDIO = "Misc Audio"
    CD = "CD"
    HOME_CONTROL = "Home Control"
    RUSSOUND_MEDIA_STREAMER = "Russound Media Streamer"
    RUSSOUND_DMS_3_1_AM_FM_TUNER = "Russound DMS 3.1 AM/FM Tuner"
    RUSSOUND_ST_1_AM_FM_TUNER = "Russound ST.1 AM/FM Tuner"
    RUSSOUND_BLUETOOTH_MODULE = "Russound Bluetooth Module"


class SourceMode(StrEnum):
    """Russound source modes."""

    UNKNOWN = "Unknown"
    AIRPLAY = "AirPlay"
    SPOTIFY = "Spotify"
    PANDORA = "Pandora"
    SIRIUS_XM = "SiriusXM"
    TUNE_IN = "TuneIn"
    INTERNET_RADIO = "Internet Radio"
    MEDIA_SERVER = "Media Server"
    USB = "USB"
    AIRABLE_RADIO = "Airable Radio"
    DEEZER = "Deezer"
    TIDAL = "Tidal"
    NAPSTER = "Napster"
    CHROMECAST = "Chromecast"
    BLUETOOTH = "Bluetooth"


class RepeatMode(StrEnum):
    """Repeat mode."""

    OFF = "OFF"
    ALL = "ALL"
    SINGLE = "SINGLE"


class PlayStatus(StrEnum):
    """Play status"""

    PLAYING = "playing"
    PAUSED = "paused"
    STOPPED = "stopped"
    TRANSITIONING = "transitioning"


@dataclass
class Source(DataClassORJSONMixin):
    """Data class representing Russound source."""

    name: str = field(default=None)
    type: SourceType = field(
        metadata={"deserialize": lambda v: SourceType.MISC_AUDIO if not v else v},
        default=SourceType.MISC_AUDIO,
    )
    channel: Optional[str] = field(default=None)
    cover_art_url: Optional[str] = field(
        metadata=field_options(alias="coverArtURL"), default=None
    )
    channel_name: Optional[str] = field(
        metadata=field_options(alias="channelName"), default=None
    )
    genre: Optional[str] = field(default=None)
    artist_name: Optional[str] = field(
        metadata=field_options(alias="artistName"), default=None
    )
    album_name: Optional[str] = field(
        metadata=field_options(alias="albumName"), default=None
    )
    playlist_name: Optional[str] = field(
        metadata=field_options(alias="playlistName"), default=None
    )
    song_name: Optional[str] = field(
        metadata=field_options(alias="songName"), default=None
    )
    program_service_name: Optional[str] = field(
        metadata=field_options(alias="programServiceName"), default=None
    )
    radio_text: Optional[str] = field(
        metadata=field_options(alias="radioText"), default=None
    )
    shuffle_mode: bool = field(
        metadata=field_options(
            alias="shuffleMode", serialization_strategy=RussoundBool()
        ),
        default=False,
    )
    repeat_mode: Optional[RepeatMode] = field(
        metadata=field_options(alias="repeatMode"), default=None
    )
    mode: SourceMode = field(
        metadata={"deserialize": lambda v: SourceMode.UNKNOWN if not v else v},
        default=SourceMode.UNKNOWN,
    )
    play_status: Optional[PlayStatus] = field(
        metadata=field_options(alias="playStatus"), default=None
    )
    sample_rate: Optional[int] = field(
        metadata=field_options(
            alias="sampleRate", serialization_strategy=RussoundInt()
        ),
        default=None,
    )
    bit_rate: Optional[int] = field(
        metadata=field_options(alias="bitRate", serialization_strategy=RussoundInt()),
        default=None,
    )
    bit_depth: Optional[int] = field(
        metadata=field_options(alias="bitDepth", serialization_strategy=RussoundInt()),
        default=None,
    )
    play_time: Optional[int] = field(
        metadata=field_options(alias="playTime", serialization_strategy=RussoundInt()),
        default=None,
    )
    track_time: Optional[int] = field(
        metadata=field_options(alias="trackTime", serialization_strategy=RussoundInt()),
        default=None,
    )
    position_last_updated: datetime = datetime.now(UTC)



class CallbackType(StrEnum):
    """Callback type."""

    STATE = "state"
    CONNECTION = "connection"


class MessageType(StrEnum):
    """Message type."""

    STATE = "S"
    NOTIFICATION = "N"
    ERROR = "E"


@dataclass
class RussoundMessage:
    """Incoming russound message."""

    type: str
    branch: Optional[str] = None
    leaf: Optional[str] = None
    value: Optional[str] = None



"""
.. include:: ../README.md
"""

from .exceptions import (
    RussoundError,
    CommandError,
    UncachedVariableError,
    UnsupportedFeatureError,
    UnsupportedRussoundVersionError,
)
from .connection import RussoundTcpConnectionHandler
from .models import Source, RussoundMessage, Zone
from .rio import Controller, RussoundClient

__all__ = [
    "RussoundError",
    "CommandError",
    "UnsupportedFeatureError",
    "UnsupportedRussoundVersionError",
    "UncachedVariableError",
    "RussoundClient",
    "Controller",
    "Zone",
    "RussoundTcpConnectionHandler",
    "Source",
    "RussoundMessage",
]



"""Asynchronous Python client for Russound RIO."""

import re

from aiorussound.const import VERSIONS_BY_FLAGS, FeatureFlag
from aiorussound.exceptions import UnsupportedFeatureError

_fw_pattern = re.compile(r"^(?P<major>\d{1,2})\.(?P<minor>\d{2})\.(?P<patch>\d{2})$")


def raise_unsupported_feature(api_ver: str, flag: FeatureFlag) -> None:
    """Raise an UnsupportedFeature exception if the specified feature is not supported
    in the provided version.
    """
    if not is_feature_supported(api_ver, flag):
        err = f"Russound feature {flag} not supported in api v{api_ver}"
        raise UnsupportedFeatureError(err)


def is_feature_supported(api_ver: str, flag: FeatureFlag) -> bool:
    """Return true if the feature is supported in the provided version,
    false otherwise.
    """
    return is_fw_version_higher(api_ver, VERSIONS_BY_FLAGS[flag])


def is_fw_version_higher(fw_a: str, fw_b: str) -> bool:
    """Return true if fw_a is greater than or equal to fw_b."""
    a_match = _fw_pattern.match(fw_a)
    b_match = _fw_pattern.match(fw_b)

    if not (a_match and b_match):
        return False

    a_major = int(a_match.group("major"))
    a_minor = int(a_match.group("minor"))
    a_patch = int(a_match.group("patch"))

    b_major = int(b_match.group("major"))
    b_minor = int(b_match.group("minor"))
    b_patch = int(b_match.group("patch"))

    return (
        (a_major > b_major)
        or (a_major == b_major and a_minor > b_minor)
        or (a_major == b_major and a_minor == b_minor and a_patch >= b_patch)
    )


def controller_device_str(controller_id: int) -> str:
    """Return a string representation of the specified controller device."""
    return f"C[{controller_id}]"


def zone_device_str(controller_id: int, zone_id: int) -> str:
    """Return a string representation of the specified zone device."""
    return f"C[{controller_id}].Z[{zone_id}]"


def source_device_str(source_id: int) -> str:
    """Return a string representation of the specified source device."""
    return f"S[{source_id}]"


def get_max_zones(model: str) -> int:
    """Return a maximum number of zones supported by a specific controller."""
    if model in ("MCA-88", "MCA-88X", "MCA-C5"):
        return 8
    if model in ("MCA-66", "MCA-C3"):
        return 6
    return 1


def is_rnet_capable(model: str) -> bool:
    """Return whether a controller is rnet capable."""
    return model in ("MCA-88X", "MCA-88", "MCA-66", "MCA-C5", "MCA-C3")


def map_rio_to_dict(state: dict, branch: str, leaf: str, value: str) -> None:
    """Maps a RIO variable to a python dictionary."""
    path = re.findall(r"\w+\[?\d*]?", branch)
    current = state
    for part in path:
        match = re.match(r"(\w+)\[(\d+)]", part)
        if match:
            key, index = match.groups()
            index = int(index)
            if key not in current:
                current[key] = {}
            if index not in current[key]:
                current[key][index] = {}
            current = current[key][index]
        else:
            if part not in current:
                current[part] = {}
            current = current[part]

    # Set the leaf and value in the final dictionary location
    current[leaf] = value



import asyncio
import logging
from abc import abstractmethod
from asyncio import StreamReader
from typing import Optional

from aiorussound.const import (
    DEFAULT_PORT,
    TIMEOUT,
)

_LOGGER = logging.getLogger(__package__)


class RussoundConnectionHandler:
    def __init__(self) -> None:
        self.reader: Optional[StreamReader] = None

    async def send(self, cmd: str) -> None:
        """Send a command to the Russound client."""
        pass
        # if not self.connected:
        #     raise CommandError("Not connected to device.")

    @abstractmethod
    async def connect(self) -> None:
        raise NotImplementedError


class RussoundTcpConnectionHandler(RussoundConnectionHandler):
    def __init__(self, host: str, port: int = DEFAULT_PORT) -> None:
        """Initialize the Russound object using the event loop, host and port
        provided.
        """
        super().__init__()
        self.host = host
        self.port = port
        self.writer = None

    async def connect(self) -> None:
        _LOGGER.debug("Connecting to %s:%s", self.host, self.port)
        async with asyncio.timeout(TIMEOUT):
            reader, writer = await asyncio.open_connection(self.host, self.port)
        self.reader = reader
        self.writer = writer

    async def send(self, cmd: str) -> None:
        """Send a command to the Russound client."""
        await super().send(cmd)
        self.writer.write(bytearray(f"{cmd}\r", "utf-8"))
        await self.writer.drain()
